@@ -7,6 +7,9 @@ Licensed under Simplified BSD License (see LICENSE)
 (C) Hiveary, LLC 2013 all rights reserved
 """
 
+import glob
+import impala
+import importlib
 import inspect
 import json
 import logging
@@ -67,11 +70,6 @@ class RealityAuditor(daemon.Daemon):
       if not os.path.isdir(self.monitors_dir):
         os.makedirs(self.monitors_dir)
 
-    self.monitor_config = stored_config.get('monitors')
-    if self.monitor_config is None:
-      self.monitor_config = {'resources': ['ResourceMonitor']}
-      parsed_args['update'] = True
-
     self.monitors = []
 
     # Get and possibly save optional configuration parameters. If the defaults
@@ -106,7 +104,7 @@ class RealityAuditor(daemon.Daemon):
       # Setup the auto-updater
       update_path = 'https://{host}/versions'.format(host=self.network_controller.remote_host)
       self.app = esky.Esky(sys.executable, update_path)
-      reactor.callLater(0, self.start_loop, self.UPDATE_TIMER, True,
+      reactor.callLater(0, self.start_loop, self.UPDATE_TIMER,
                         self.auto_agent_update)
 
     # Load all monitors
@@ -139,7 +137,7 @@ class RealityAuditor(daemon.Daemon):
 
     # Send a ping to the server to act as a keep-alive.
     reactor.callLater(self.INITIAL_DELAY, self.start_loop,
-                      self.network_controller.PING_TIMER, True,
+                      self.network_controller.PING_TIMER,
                       self.network_controller.ping_pong)
 
     reactor.run()
@@ -150,7 +148,7 @@ class RealityAuditor(daemon.Daemon):
     """
 
     # Check that we have monitors enabled in the config, and know where to find them
-    if self.monitor_config and self.monitors_dir:
+    if self.monitors_dir:
       # Add the monitors dir to path so we can import monitors
       self.logger.info('Loading monitors from %s', self.monitors_dir)
       sys.path.insert(0, self.monitors_dir)
@@ -164,22 +162,33 @@ class RealityAuditor(daemon.Daemon):
         self.logger.debug('Also loading frozen monitors from: %s',
                           frozen_monitors_dir)
 
+      # Find all of the monitor modules.
+      monitor_map = {}
+      for filename in glob.glob('%s/*.py' % self.monitors_dir):
+        monitor = os.path.basename(filename).split('.')[0]
+        monitor_map['hiveary.monitors.%s' % monitor] = filename
+
       # Import all of the monitors and add the instances to our monitor list.
-      for module_name, monitor_classes in self.monitor_config.iteritems():
+      impala.register(monitor_map)
+      object_filter = lambda obj: inspect.isclass(obj) and monitors.BaseMonitor in inspect.getmro(obj)
+      for module_name in monitor_map.keys():
         try:
-          self.logger.info('Loading monitor %s from file %s', monitor_classes, module_name)
-          module = __import__(module_name, globals(), locals(),
-                              fromlist=monitor_classes)
-          for class_name in monitor_classes:
-            monitor_class = getattr(module, class_name)
-            # Make sure this class inherits the monitors.BaseMonitor class.
-            if monitors.BaseMonitor in inspect.getmro(monitor_class):
-              monitor = monitor_class()
-              self.monitors.append(monitor)
-            else:
-              self.logger.warn('Tried to load %s, but was not a HivearyMonitor', monitor_class)
-        except:
-          self.logger.error('Failed to load module %s', module_name)
+          module = importlib.import_module(module_name)
+        except ImportError:
+          self.logger.error('Failed to import module %s', module_name)
+          continue
+
+        # Filter down to all classes that inherit the monitors.BaseMonitor class.
+        for class_name, monitor_class in inspect.getmembers(module, object_filter):
+          self.logger.info('Loading %s from %s', class_name, module_name)
+          try:
+            monitor = monitor_class()
+          except Exception:
+            self.logger.error('Failed to load class %s from module %s',
+                              class_name, module_name)
+            self.logger.debug('Full loading error:', exc_info=True)
+          else:
+            self.monitors.append(monitor)
 
   def start_monitor(self, monitor):
     """Starts a given monitor.
@@ -190,12 +199,12 @@ class RealityAuditor(daemon.Daemon):
 
     self.logger.debug('Starting %s (%s) monitor data checks', monitor.NAME,
                       monitor.UID)
-    self.network_controller.expected_values[monitor.UID] = monitor.expected_values
+    self.network_controller.monitors[monitor.UID] = monitor
     monitor.send_alert = self.network_controller.publish_alert_message
 
     # Check if the monitor should run in a loop
     if monitor.MONITOR_TIMER is not None:
-      self.start_loop(monitor.MONITOR_TIMER, False, monitor.run)
+      self.start_loop(monitor.MONITOR_TIMER, monitor.run)
     else:
       reactor.callInThread(monitor.run)
 
@@ -210,13 +219,17 @@ class RealityAuditor(daemon.Daemon):
     """
 
     self.logger.info("Received signal: %s", signum)
+    self.shutdown()
 
-    self.network_controller.stop_amqp()
-    reactor.callFromThread(reactor.stop)
+  def shutdown(self):
+    """Stops the agent and any active network connections."""
 
     # Clean up the daemon after the reactor is done
     reactor.addSystemEventTrigger('after', 'shutdown', self.delpid)
 
+    self.network_controller.stop_amqp()
+    reactor.stop()
+
   def set_config(self, args, stored_config):
     """Update any configuration variables. Use passed command-line values first,
     then stored config values, then defaults.
@@ -279,28 +292,24 @@ class RealityAuditor(daemon.Daemon):
       # period if we have data points from at least half that period.
       delta = monitor.next_interval()
       self.logger.debug('Starting %s aggregation loop in %s seconds', monitor.NAME, delta)
-      start_now = delta > (monitor.AGGREGATION_TIMER / 2)
     else:
       delta = monitor.AGGREGATION_TIMER
-      start_now = True
 
     reactor.callLater(int(delta), self.start_loop,
-                      monitor.AGGREGATION_TIMER, start_now,
+                      monitor.AGGREGATION_TIMER,
                       monitor.send_data, self.network_controller)
 
-  def start_loop(self, timer, start_now, func, *args, **kwargs):
+  def start_loop(self, timer, func, *args, **kwargs):
     """Continuously loop the passed function.
 
     Args:
       timer: How often to loop the call, in seconds.
-      start_now: Boolean indicating whether the first iteration should happen
-                 immediately or wait until after the first timer expires.
       func: The function to run.
       *args, **kwargs: Anything that needs to be passed to the function.
     """
 
     loop = task.LoopingCall(func, *args, **kwargs)
-    deferred_task = loop.start(timer, now=start_now)
+    deferred_task = loop.start(timer, now=True)
     deferred_task.addErrback(self.logger.error)
 
   def restart(self):
@@ -308,12 +317,9 @@ class RealityAuditor(daemon.Daemon):
 
     self.logger.warn('The agent is restarting...')
 
-    reactor.addSystemEventTrigger('after', 'shutdown', self.delpid)
     reactor.addSystemEventTrigger('after', 'shutdown', self.fork,
                                   detached=False, exit=False)
-
-    self.network_controller.stop_amqp()
-    reactor.stop()
+    self.shutdown()
 
   def auto_agent_update(self):
     """Checks for a new version of the running application from the remote server.
