@@ -11,6 +11,7 @@ import datetime
 import json
 import kombu
 import kombu.common
+import kombu.utils.debug
 import logging
 import oauth2
 import os
@@ -73,7 +74,7 @@ class NetworkController(object):
     # Thread and deferred management
     self.reactor = reactor
 
-    self.expected_values = {}
+    self.monitors = {}
 
   def ensure_internet_connection(self, test_url='http://198.41.189.27'):
     """Blocks until there is an active connection to the public internet.
@@ -121,6 +122,9 @@ class NetworkController(object):
       self.logger.error('Missing required parameters to establish an AMQP connection')
       sys.exit(1)
 
+    if self.debug_mode:
+      kombu.utils.debug.setup_logging()
+
     # Create the connection
     ca_certs = os.path.join(paths.get_program_path(), 'ca-bundle.pem')
     self.logger.debug('Using SSL cert bundle at "%s"', ca_certs)
@@ -136,7 +140,8 @@ class NetworkController(object):
 
     self.logger.debug('Connecting to %s as user %s', self.amqp_server, self.user_id)
     self.amqp = kombu.Connection(self.amqp_server, self.user_id, self.amqp_password,
-                                 port=5671, ssl=ssl_options, insist=True)
+                                 port=5671, ssl=ssl_options, insist=True,
+                                 transport='amqplib')
 
     self.amqp.ensure_connection(errback=self.amqp_errback, interval_max=60)
     self.logger.info('SSL-AMQP connection established')
@@ -158,6 +163,7 @@ class NetworkController(object):
     the error, so ensure_connection will not be enoguh to reconnnect.
     Forcing a release first resolves this."""
 
+    self.logger.info('Reconnecting to AMQP...')
     self.amqp.release()
     self.amqp.ensure_connection(errback=self.amqp_errback,
                                 interval_start=5,
@@ -180,19 +186,22 @@ class NetworkController(object):
       if self.running:
         raise
 
-    loop = kombu.common.eventloop(self.amqp, timeout=1, ignore_timeouts=True)
     self.logger.info('Draining events from the server')
 
     while self.running:
       try:
-        next(loop)
+        self.amqp.drain_events()
+      except (socket.timeout, socket.error):
+        pass
       except Exception, err:
         # Errors generated while the agent is stopping can be ignored
         if self.running:
-          self.logger.error('AMQP error while draining events: %s', err)
+          self.logger.error('AMQP error while draining events: %s', err,
+                            exc_info=self.debug_mode)
           self.amqp_reconnect()
           break
     else:
+      self.logger.debug('No longer running, stopping the amqp drain')
       return
 
     # If the loop broke while the agent is still running, try starting over
@@ -229,8 +238,15 @@ class NetworkController(object):
       A tuple of (response, content), the first being an instance of the
       'httplib2.Http.Response' class, the second being a string that
       contains the response entity body.
+
+    Raises:
+      RuntimeError: Occurs when the program is exiting but we are stuck in this loop.
     """
 
+    # Check to see if we tried to kill the process, if so, bail.
+    if not self.running:
+      raise RuntimeError
+
     client = self.create_oauth_client()
 
     try:
@@ -273,32 +289,44 @@ class NetworkController(object):
           and will be JSONified before being sent.
     """
 
-    alert['host_id'] = self.obj_id
-    self.publish_info_message('alert', json.dumps(alert))
+    self.publish_info_message('alert', alert)
 
-  def publish_info_message(self, destination, message='', retry=True):
+  def publish_info_message(self, routing_key, message='', retry=True,
+                           exchange_name=None):
     """Method to publish an AMQP message.
 
     Args:
-      destination: The AMQP routing key.
-      message: The message to publish, as a string (usually a JSONified object).
+      routing_key: The AMQP routing key.
+      message: The message to publish, as a string or dict.
+      retry: A boolean of whether a failed publish should be re-attempted.
+      exchange_name: An optional AMQP exchange name to use.
     """
 
-    self.logger.debug('Sending "%s" AMQP message', destination)
+    self.logger.debug('Sending "%s" AMQP message', routing_key)
 
-    exchange = kombu.Exchange('agent.{user}'.format(user=self.user_id))
-    with self.amqp.Producer(exchange=exchange, routing_key=destination,
+    if not exchange_name:
+      exchange_name = 'agent.{user}'.format(user=self.user_id)
+
+    # Check the message format and add the host ID
+    if type(message) == dict:
+      message['host_id'] = self.obj_id
+      message['timestamp'] = time.time()
+      message = json.dumps(message)
+
+    exchange = kombu.Exchange(exchange_name)
+    with self.amqp.Producer(exchange=exchange, routing_key=routing_key,
                             auto_declare=False) as producer:
       try:
         producer.publish(message, user_id=self.user_id,
                          timestamp=datetime.datetime.utcnow())
       except Exception, err:
-        self.logger.error('Error while publishing to AMQP: %s', err)
+        self.logger.error('Error while publishing to AMQP: %s', err,
+                          exc_info=self.debug_mode)
         self.amqp_reconnect()
 
         # Retry publishing the message if requested
         if retry:
-          self.publish_info_message(destination, message)
+          self.publish_info_message(routing_key, message, retry, exchange_name)
 
   def amqp_errback(self, exc, interval):
     """Error callback fired when there is a problem with the connection or channel.
@@ -315,9 +343,8 @@ class NetworkController(object):
     """Function to alert the server that we're still alive and doing science."""
 
     # Send the ping
-    data = {'host_id': self.obj_id}
-    self.logger.debug('Sending ping to server with a body of: %s', data)
-    self.publish_info_message('ping', json.dumps(data), retry=False)
+    self.logger.debug('Sending ping to server')
+    self.publish_info_message('ping', {}, retry=False)
 
   def task_callback(self, body, message):
     """Callback for when a message is received from the tasks queue.
@@ -328,6 +355,7 @@ class NetworkController(object):
     """
 
     self.logger.debug('Received the following task message: %s', body)
+    message.ack()
 
     try:
       data = json.loads(body)
@@ -336,9 +364,6 @@ class NetworkController(object):
     else:
       self.run_task(data)
 
-    # Let the broker know that we're done processing this message
-    message.ack()
-
   def run_task(self, client_task):
     """Run a task as commanded by the control server.
 
@@ -395,8 +420,51 @@ class NetworkController(object):
       self.logger.info('Received new %s expected values: %s',
                        monitor_id, expected_values)
 
-      if monitor_id in self.expected_values:
-        self.expected_values[monitor_id].update(expected_values)
+      if monitor_id in self.monitors:
+        self.monitors[monitor_id].expected_values.update(expected_values)
+      else:
+        self.logger.warn('Monitor "%s" is not enabled!', monitor_id)
+    elif task_name == 'live_data':
+      # Tell the relevant monitor to start sending a copy of all data to a special
+      # real-time AMQP queue
+      monitor_id = client_task['command']['monitor']
+      action = client_task['command']['action']
+      stream_routing_key = client_task['command']['routing_key']
+      self.logger.info('Received request to %s sending real-time data for %s',
+                       action, monitor_id)
+
+      if monitor_id in self.monitors:
+        if action == 'start':
+          # Add a new livestream callback
+          exchange_name = 'agent.{user}.reports'.format(user=self.user_id)
+          live_publish = lambda data: self.publish_info_message(stream_routing_key,
+                                                                data,
+                                                                exchange_name=exchange_name)
+
+          # Send a copy of any data that has been aggregated so far
+          data = self.monitors[monitor_id].merge_data()
+          data.pop('timestamp', None)
+          data_container = {
+              'data': data,
+              'monitor_id': monitor_id,
+              'interval': self.monitors[monitor_id].MONITOR_TIMER,
+          }
+          self.logger.debug('Sending inititial data: %s', data_container)
+          live_publish(data_container)
+
+          # Store the lambda on the monitor so all future data will get published
+          self.monitors[monitor_id].livestreams[stream_routing_key] = live_publish
+        elif action == 'stop':
+          # Delete the previously setup stream
+          self.logger.info('Stopping livestream for %s...', monitor_id)
+          if monitor_id not in self.monitors:
+            self.logger.warn('Monitor %s has not been started', monitor_id)
+          elif stream_routing_key not in self.monitors[monitor_id].livestreams:
+            self.logger.info('Livestream for monitor %s was not enabled, skipping',
+                             monitor_id)
+          else:
+            del(self.monitors[monitor_id].livestreams[stream_routing_key])
+            self.logger.info('Livestream stopped')
       else:
         self.logger.warn('Monitor "%s" is not enabled!', monitor_id)
     else:
